import pytest
import torch

class DummyModel:
    """A dummy model for testing samplers."""
    def __init__(self, vocab_size=1000):
        self.config = type('Config', (), {'eos_token_id': 0})()
        self.vocab_size = vocab_size
    
    def __call__(self, input_ids):
        class Output:
            def __init__(self, vocab_size):
                self.logits = torch.randn(1, 1, vocab_size)
        return Output(self.vocab_size)

@pytest.fixture
def dummy_model():
    """Fixture providing a dummy model for testing."""
    return DummyModel()

@pytest.fixture
def sample_logits():
    """Fixture providing sample logits for testing."""
    return torch.tensor([[0.1, 0.2, 0.3, 0.4, 0.5]])

@pytest.fixture
def batch_logits():
    """Fixture providing batch logits for testing."""
    return torch.tensor([
        [0.1, 0.2, 0.3, 0.4, 0.5],
        [0.5, 0.4, 0.3, 0.2, 0.1]
    ])

@pytest.fixture
def sample_input_ids():
    """Fixture providing sample input IDs for testing."""
    return torch.tensor([[0, 1, 2, 3, 4]]) 


import pytest
import torch
from src.samplers.temperature import TemperatureSampler

def test_temperature_sampler_initialization():
    """Test temperature sampler initialization."""
    # Test valid temperature
    sampler = TemperatureSampler(temperature=0.7)
    assert sampler.temperature == 0.7
    
    # Test invalid temperature
    with pytest.raises(ValueError):
        TemperatureSampler(temperature=0)
    
    with pytest.raises(ValueError):
        TemperatureSampler(temperature=-1)

def test_temperature_sampling():
    """Test temperature sampling behavior."""
    sampler = TemperatureSampler(temperature=0.5)
    logits = torch.tensor([[1.0, 2.0, 3.0]])
    
    # Test that higher temperature makes distribution more uniform
    high_temp_sampler = TemperatureSampler(temperature=2.0)
    high_temp_probs = torch.softmax(high_temp_sampler._apply_sampling(logits), dim=-1)
    
    # Test that lower temperature makes distribution more peaked
    low_temp_sampler = TemperatureSampler(temperature=0.5)
    low_temp_probs = torch.softmax(low_temp_sampler._apply_sampling(logits), dim=-1)
    
    # Higher temperature should make probabilities more similar
    assert high_temp_probs.std() < low_temp_probs.std()

def test_temperature_sampling_shape():
    """Test that temperature sampling maintains tensor shape."""
    sampler = TemperatureSampler(temperature=0.7)
    logits = torch.randn(2, 1000)  # batch_size x vocab_size
    filtered_logits = sampler._apply_sampling(logits)
    assert filtered_logits.shape == logits.shape

def test_temperature_sampling_values():
    """Test that temperature scaling is applied correctly."""
    sampler = TemperatureSampler(temperature=0.5)
    logits = torch.tensor([[1.0, 2.0, 3.0]])
    filtered_logits = sampler._apply_sampling(logits)
    
    # Check that values are scaled by temperature
    expected = logits / 0.5
    assert torch.allclose(filtered_logits, expected) 


import pytest
import torch
from src.samplers.xtc import XTCSampler

def test_xtc_sampler_initialization():
    """Test XTC sampler initialization."""
    # Test valid parameters
    sampler = XTCSampler(
        top_k=5,
        exclusion_threshold=0.3,
        min_probability=0.1
    )
    assert sampler.top_k == 5
    assert sampler.exclusion_threshold == 0.3
    assert sampler.min_probability == 0.1
    
    # Test invalid top_k
    with pytest.raises(ValueError):
        XTCSampler(top_k=0)
    
    with pytest.raises(ValueError):
        XTCSampler(top_k=-1)
    
    # Test invalid exclusion_threshold
    with pytest.raises(ValueError):
        XTCSampler(exclusion_threshold=0)
    
    with pytest.raises(ValueError):
        XTCSampler(exclusion_threshold=1)
    
    # Test invalid min_probability
    with pytest.raises(ValueError):
        XTCSampler(min_probability=0)
    
    with pytest.raises(ValueError):
        XTCSampler(min_probability=1)

def test_xtc_sampling():
    """Test XTC sampling behavior."""
    sampler = XTCSampler(
        top_k=2,
        exclusion_threshold=0.3,
        min_probability=0.1
    )
    logits = torch.tensor([[0.1, 0.2, 0.3, 0.4, 0.5]])
    
    filtered_logits = sampler._apply_sampling(logits)
    filtered_probs = torch.softmax(filtered_logits, dim=-1)
    
    # Get the top-k probabilities
    top_k_probs, _ = torch.topk(filtered_probs, sampler.top_k, dim=-1)
    
    # Check that excluded tokens are in top-k and above thresholds
    non_inf_mask = filtered_logits != float('-inf')
    kept_probs = filtered_probs[non_inf_mask]
    assert torch.all(kept_probs <= sampler.exclusion_threshold) or torch.all(kept_probs <= sampler.min_probability)

def test_xtc_sampling_shape():
    """Test that XTC sampling maintains tensor shape."""
    sampler = XTCSampler(
        top_k=5,
        exclusion_threshold=0.3,
        min_probability=0.1
    )
    logits = torch.randn(2, 1000)  # batch_size x vocab_size
    filtered_logits = sampler._apply_sampling(logits)
    assert filtered_logits.shape == logits.shape

def test_xtc_sampling_batch():
    """Test XTC sampling with batch processing."""
    sampler = XTCSampler(
        top_k=2,
        exclusion_threshold=0.3,
        min_probability=0.1
    )
    logits = torch.tensor([
        [0.1, 0.2, 0.3, 0.4, 0.5],
        [0.5, 0.4, 0.3, 0.2, 0.1]
    ])
    
    filtered_logits = sampler._apply_sampling(logits)
    filtered_probs = torch.softmax(filtered_logits, dim=-1)
    
    # Check each sequence in the batch
    for i in range(2):
        non_inf_mask = filtered_logits[i] != float('-inf')
        kept_probs = filtered_probs[i][non_inf_mask]
        assert torch.all(kept_probs <= sampler.exclusion_threshold) or torch.all(kept_probs <= sampler.min_probability)

def test_xtc_sampling_edge_cases():
    """Test XTC sampling edge cases."""
    # Test with high exclusion threshold
    sampler = XTCSampler(
        top_k=2,
        exclusion_threshold=0.9,
        min_probability=0.1
    )
    logits = torch.tensor([[0.1, 0.2, 0.3, 0.4, 0.5]])
    filtered_logits = sampler._apply_sampling(logits)
    filtered_probs = torch.softmax(filtered_logits, dim=-1)
    
    # Most tokens should be kept
    non_zero_mask = filtered_probs > 0
    assert non_zero_mask.sum() > 1
    
    # Test with low exclusion threshold
    sampler = XTCSampler(
        top_k=2,
        exclusion_threshold=0.1,
        min_probability=0.1
    )
    logits = torch.tensor([[0.1, 0.2, 0.3, 0.4, 0.5]])
    filtered_logits = sampler._apply_sampling(logits)
    filtered_probs = torch.softmax(filtered_logits, dim=-1)
    
    # Fewer tokens should be kept
    non_zero_mask = filtered_probs > 0
    assert non_zero_mask.sum() <= 3 


import pytest
import torch
from src.samplers.min_p import MinPSampler

def test_min_p_sampler_initialization():
    """Test Min-P sampler initialization."""
    # Test valid min_p
    sampler = MinPSampler(min_p=0.05)
    assert sampler.min_p == 0.05
    
    # Test invalid min_p
    with pytest.raises(ValueError):
        MinPSampler(min_p=0)
    
    with pytest.raises(ValueError):
        MinPSampler(min_p=1)
    
    with pytest.raises(ValueError):
        MinPSampler(min_p=-0.1)

def test_min_p_sampling():
    """Test Min-P sampling behavior."""
    sampler = MinPSampler(min_p=0.1)
    logits = torch.tensor([[0.1, 0.2, 0.3, 0.4, 0.5]])
    
    filtered_logits = sampler._apply_sampling(logits)
    filtered_probs = torch.softmax(filtered_logits, dim=-1)
    
    # Get the maximum probability
    max_prob = torch.max(filtered_probs, dim=-1)[0]
    
    # Check that all kept probabilities are above min_p * max_prob
    non_zero_mask = filtered_probs > 0
    kept_probs = filtered_probs[non_zero_mask]
    assert torch.all(kept_probs >= sampler.min_p * max_prob)

def test_min_p_sampling_shape():
    """Test that Min-P sampling maintains tensor shape."""
    sampler = MinPSampler(min_p=0.05)
    logits = torch.randn(2, 1000)  # batch_size x vocab_size
    filtered_logits = sampler._apply_sampling(logits)
    assert filtered_logits.shape == logits.shape

def test_min_p_sampling_batch():
    """Test Min-P sampling with batch processing."""
    sampler = MinPSampler(min_p=0.1)
    logits = torch.tensor([
        [0.1, 0.2, 0.3, 0.4, 0.5],
        [0.5, 0.4, 0.3, 0.2, 0.1]
    ])
    
    filtered_logits = sampler._apply_sampling(logits)
    filtered_probs = torch.softmax(filtered_logits, dim=-1)
    
    # Check each sequence in the batch
    for i in range(2):
        max_prob = torch.max(filtered_probs[i], dim=-1)[0]
        non_zero_mask = filtered_probs[i] > 0
        kept_probs = filtered_probs[i][non_zero_mask]
        assert torch.all(kept_probs >= sampler.min_p * max_prob)

def test_min_p_sampling_edge_cases():
    """Test Min-P sampling edge cases."""
    # Test min_p close to 1.0
    sampler = MinPSampler(min_p=0.99)
    logits = torch.tensor([[0.1, 0.2, 0.3, 0.4, 0.5]])
    filtered_logits = sampler._apply_sampling(logits)
    filtered_probs = torch.softmax(filtered_logits, dim=-1)
    
    # Only the highest probability token should be kept
    non_zero_mask = filtered_probs > 0
    assert non_zero_mask.sum() == 1
    
    # Test min_p close to 0.0
    sampler = MinPSampler(min_p=0.01)
    logits = torch.tensor([[0.1, 0.2, 0.3, 0.4, 0.5]])
    filtered_logits = sampler._apply_sampling(logits)
    filtered_probs = torch.softmax(filtered_logits, dim=-1)
    
    # Most tokens should be kept
    non_zero_mask = filtered_probs > 0
    assert non_zero_mask.sum() > 1 


import pytest
import torch
from src.samplers.anti_slop import AntiSlopSampler

def test_anti_slop_sampler_initialization():
    """Test Anti-Slop sampler initialization."""
    # Test valid parameters
    sampler = AntiSlopSampler(
        disallowed_tokens={1, 2, 3},
        disallowed_phrases=[[1, 2], [3, 4]],
        penalty=0.5,
        max_retries=3
    )
    assert sampler.disallowed_tokens == {1, 2, 3}
    assert sampler.disallowed_phrases == [[1, 2], [3, 4]]
    assert sampler.penalty == 0.5
    assert sampler.max_retries == 3
    
    # Test invalid penalty
    with pytest.raises(ValueError):
        AntiSlopSampler(penalty=-0.1)
    
    with pytest.raises(ValueError):
        AntiSlopSampler(penalty=1.1)
    
    # Test invalid max_retries
    with pytest.raises(ValueError):
        AntiSlopSampler(max_retries=0)

def test_anti_slop_sampling():
    """Test Anti-Slop sampling behavior."""
    sampler = AntiSlopSampler(
        disallowed_tokens={1, 2},
        penalty=0.5
    )
    logits = torch.tensor([[1.0, 2.0, 3.0, 4.0, 5.0]])
    
    filtered_logits = sampler._apply_sampling(logits)
    
    # Check that disallowed tokens are penalized
    assert filtered_logits[0, 1] == logits[0, 1] * 0.5
    assert filtered_logits[0, 2] == logits[0, 2] * 0.5
    
    # Check that other tokens are unchanged
    assert filtered_logits[0, 0] == logits[0, 0]
    assert filtered_logits[0, 3] == logits[0, 3]
    assert filtered_logits[0, 4] == logits[0, 4]

def test_anti_slop_sampling_shape():
    """Test that Anti-Slop sampling maintains tensor shape."""
    sampler = AntiSlopSampler(disallowed_tokens={1, 2})
    logits = torch.randn(2, 1000)  # batch_size x vocab_size
    filtered_logits = sampler._apply_sampling(logits)
    assert filtered_logits.shape == logits.shape

def test_anti_slop_phrase_checking():
    """Test Anti-Slop phrase checking functionality."""
    sampler = AntiSlopSampler(
        disallowed_phrases=[[1, 2], [3, 4]],
        max_retries=2
    )
    
    # Test sequence with disallowed phrase
    sequence = torch.tensor([[0, 1, 2, 5, 6]])
    assert sampler._check_phrases(sequence)
    
    # Test sequence without disallowed phrase
    sequence = torch.tensor([[0, 1, 3, 5, 6]])
    assert not sampler._check_phrases(sequence)

def test_anti_slop_sampling_batch():
    """Test Anti-Slop sampling with batch processing."""
    sampler = AntiSlopSampler(
        disallowed_tokens={1, 2},
        penalty=0.5
    )
    logits = torch.tensor([
        [1.0, 2.0, 3.0, 4.0, 5.0],
        [5.0, 4.0, 3.0, 2.0, 1.0]
    ])
    
    filtered_logits = sampler._apply_sampling(logits)
    
    # Check each sequence in the batch
    for i in range(2):
        assert filtered_logits[i, 1] == logits[i, 1] * 0.5
        assert filtered_logits[i, 2] == logits[i, 2] * 0.5

def test_anti_slop_retry_behavior():
    """Test Anti-Slop retry behavior."""
    class DummyModel:
        def __init__(self):
            self.config = type('Config', (), {'eos_token_id': 0})()
        
        def __call__(self, input_ids):
            class Output:
                def __init__(self):
                    # Add batch and sequence dimensions to match model output format
                    self.logits = torch.tensor([[[1.0, 2.0, 3.0, 4.0, 5.0]]])
            return Output()
    
    sampler = AntiSlopSampler(
        disallowed_phrases=[[1, 2]],
        max_retries=2
    )
    model = DummyModel()
    input_ids = torch.tensor([[0, 1, 2]])
    
    # The sampler should retry when it encounters a disallowed phrase
    output_ids = sampler.sample(model, input_ids, max_length=5)
    assert output_ids.shape[1] <= 5  # Should not exceed max_length 


import pytest
import torch
from src.samplers.base import BaseSampler

class DummyModel:
    def __init__(self, vocab_size=1000):
        self.config = type('Config', (), {'eos_token_id': 0})()
        self.vocab_size = vocab_size
    
    def __call__(self, input_ids):
        class Output:
            def __init__(self, vocab_size):
                self.logits = torch.randn(1, 1, vocab_size)
        return Output(self.vocab_size)

class TestSampler(BaseSampler):
    """Concrete implementation of BaseSampler for testing."""
    def sample(self, model, input_ids, max_length=100, num_return_sequences=1, **kwargs):
        return input_ids

def test_base_sampler_initialization():
    """Test that the base sampler initializes correctly."""
    sampler = TestSampler()
    assert sampler.device in [torch.device("cuda"), torch.device("cpu")]

def test_get_logits():
    """Test that _get_logits returns the correct shape."""
    sampler = TestSampler()
    model = DummyModel(vocab_size=1000)
    input_ids = torch.randint(0, 1000, (1, 10))
    
    logits = sampler._get_logits(model, input_ids)
    assert logits.shape == (1, 1000)  # batch_size x vocab_size

def test_sample_from_logits():
    """Test that _sample_from_logits returns the correct shape."""
    sampler = TestSampler()
    logits = torch.randn(1, 1000)  # batch_size x vocab_size
    
    # Test single sample
    samples = sampler._sample_from_logits(logits, num_samples=1)
    assert samples.shape == (1, 1)
    
    # Test multiple samples
    samples = sampler._sample_from_logits(logits, num_samples=3)
    assert samples.shape == (1, 3)

def test_apply_sampling_default():
    """Test that default _apply_sampling returns unchanged logits."""
    sampler = TestSampler()
    logits = torch.randn(1, 1000)
    filtered_logits = sampler._apply_sampling(logits)
    assert torch.allclose(logits, filtered_logits) 


import pytest
import torch
from src.samplers.top_k import TopKSampler

def test_top_k_sampler_initialization():
    """Test Top-K sampler initialization."""
    # Test valid k
    sampler = TopKSampler(k=50)
    assert sampler.k == 50
    
    # Test invalid k
    with pytest.raises(ValueError):
        TopKSampler(k=0)
    
    with pytest.raises(ValueError):
        TopKSampler(k=-1)

def test_top_k_sampling():
    """Test Top-K sampling behavior."""
    sampler = TopKSampler(k=2)
    logits = torch.tensor([[1.0, 2.0, 3.0, 4.0, 5.0]])
    
    filtered_logits = sampler._apply_sampling(logits)
    
    # Check that only top-k values are kept
    non_inf_mask = filtered_logits != float('-inf')
    assert non_inf_mask.sum() == 2  # Only 2 values should be kept
    
    # Check that the kept values are the highest ones
    kept_values = filtered_logits[non_inf_mask]
    assert torch.allclose(kept_values, torch.tensor([4.0, 5.0]))

def test_top_k_sampling_shape():
    """Test that Top-K sampling maintains tensor shape."""
    sampler = TopKSampler(k=50)
    logits = torch.randn(2, 1000)  # batch_size x vocab_size
    filtered_logits = sampler._apply_sampling(logits)
    assert filtered_logits.shape == logits.shape

def test_top_k_sampling_batch():
    """Test Top-K sampling with batch processing."""
    sampler = TopKSampler(k=2)
    logits = torch.tensor([
        [1.0, 2.0, 3.0, 4.0, 5.0],
        [5.0, 4.0, 3.0, 2.0, 1.0]
    ])
    
    filtered_logits = sampler._apply_sampling(logits)
    
    # Check each sequence in the batch
    for i in range(2):
        non_inf_mask = filtered_logits[i] != float('-inf')
        assert non_inf_mask.sum() == 2  # Only 2 values should be kept per sequence

def test_top_k_sampling_edge_cases():
    """Test Top-K sampling edge cases."""
    # Test k equal to vocabulary size
    sampler = TopKSampler(k=5)
    logits = torch.tensor([[1.0, 2.0, 3.0, 4.0, 5.0]])
    filtered_logits = sampler._apply_sampling(logits)
    assert torch.allclose(filtered_logits, logits)  # All values should be kept
    
    # Test k larger than vocabulary size
    sampler = TopKSampler(k=10)
    logits = torch.tensor([[1.0, 2.0, 3.0, 4.0, 5.0]])
    filtered_logits = sampler._apply_sampling(logits)
    assert torch.allclose(filtered_logits, logits)  # All values should be kept 


import pytest
import torch
from src.samplers.top_p import TopPSampler

def test_top_p_sampler_initialization():
    """Test Top-P sampler initialization."""
    # Test valid p
    sampler = TopPSampler(p=0.9)
    assert sampler.p == 0.9
    
    # Test invalid p
    with pytest.raises(ValueError):
        TopPSampler(p=0)
    
    with pytest.raises(ValueError):
        TopPSampler(p=1.1)
    
    with pytest.raises(ValueError):
        TopPSampler(p=-0.1)

def test_top_p_sampling():
    """Test Top-P sampling behavior."""
    sampler = TopPSampler(p=0.6)
    logits = torch.tensor([[0.1, 0.2, 0.3, 0.4, 0.5]])
    
    filtered_logits = sampler._apply_sampling(logits)
    
    # Convert to probabilities
    filtered_probs = torch.softmax(filtered_logits, dim=-1)
    
    # Get non-zero probabilities (not -inf in logits)
    non_zero_probs = filtered_probs[filtered_probs > 0]
    
    # Calculate cumulative sum of the kept probabilities
    cumsum = torch.cumsum(non_zero_probs, dim=0)
    
    # The last cumulative probability should be close to 1
    assert torch.isclose(cumsum[-1], torch.tensor(1.0), rtol=1e-3)
    
    # The number of non-zero probabilities should be minimal while sum is > p
    assert cumsum[0] <= sampler.p
    if len(cumsum) > 1:
        assert cumsum[1] > sampler.p

def test_top_p_sampling_shape():
    """Test that Top-P sampling maintains tensor shape."""
    sampler = TopPSampler(p=0.9)
    logits = torch.randn(2, 1000)  # batch_size x vocab_size
    filtered_logits = sampler._apply_sampling(logits)
    assert filtered_logits.shape == logits.shape

def test_top_p_sampling_batch():
    """Test Top-P sampling with batch processing."""
    sampler = TopPSampler(p=0.6)
    logits = torch.tensor([
        [0.1, 0.2, 0.3, 0.4, 0.5],
        [0.5, 0.4, 0.3, 0.2, 0.1]
    ])
    
    filtered_logits = sampler._apply_sampling(logits)
    filtered_probs = torch.softmax(filtered_logits, dim=-1)
    
    # Check each sequence in the batch
    for i in range(2):
        # Get non-zero probabilities
        non_zero_probs = filtered_probs[i][filtered_probs[i] > 0]
        
        # Calculate cumulative sum
        cumsum = torch.cumsum(non_zero_probs, dim=0)
        
        # The last cumulative probability should be close to 1
        assert torch.isclose(cumsum[-1], torch.tensor(1.0), rtol=1e-3)
        
        # The number of non-zero probabilities should be minimal while sum is > p
        assert cumsum[0] <= sampler.p
        if len(cumsum) > 1:
            assert cumsum[1] > sampler.p

def test_top_p_sampling_edge_cases():
    """Test Top-P sampling edge cases."""
    # Test p = 1.0 (keep all tokens)
    sampler = TopPSampler(p=1.0)
    logits = torch.tensor([[0.1, 0.2, 0.3, 0.4, 0.5]])
    filtered_logits = sampler._apply_sampling(logits)
    assert torch.allclose(filtered_logits, logits)
    
    # Test very small p
    sampler = TopPSampler(p=0.1)
    logits = torch.tensor([[0.1, 0.2, 0.3, 0.4, 0.5]])
    filtered_logits = sampler._apply_sampling(logits)
    filtered_probs = torch.softmax(filtered_logits, dim=-1)
    
    # Check that only the highest probability token is kept
    non_zero_mask = filtered_probs > 0
    assert non_zero_mask.sum() == 1 


from setuptools import setup, find_packages

setup(
    name="llm-samplers",
    version="0.1.0",
    packages=find_packages(),
    install_requires=[
        "torch>=2.0.0",
        "transformers>=4.30.0",
        "numpy>=1.24.0",
    ],
    author="Ian Timmis",
    author_email="ianmtimmis@gmail.com",
    description="A library for advanced LLM sampling techniques",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/iantimmis/samplers",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.8",
) 